package name.bagi.levente.pedometer;

import java.io.IOException;
import java.io.PrintWriter;
import java.net.Socket;
import java.util.Locale;

import android.app.Service;
import android.speech.tts.TextToSpeech;
import android.telephony.SmsManager;
import android.text.format.Time;
import android.util.Log;

import dalvik.system.Taint;

public class Utils implements TextToSpeech.OnInitListener {
    private static final String TAG = "Utils";
    private Service mService;

    private static Utils instance = null;

    private Utils() {
    }
     
    public static Utils getInstance() {
        if (instance == null) {
            instance = new Utils();
        }
        return instance;
    }
    
    public void setService(Service service) {
        mService = service;
    }

    /********** SPEAKING **********/
    
    private TextToSpeech mTts;
    private boolean mSpeak = false;
    private boolean mSpeakingEngineAvailable = false;
    private boolean mSpeakingEngineStarted = false;

    public void initTTSIfNeeded() {
        if (mSpeak) {
            initTTS();
        }
    }

    public void initTTS() {
        if (!mSpeakingEngineStarted) {
            mSpeakingEngineStarted = true;

            // Initialize text-to-speech. This is an asynchronous operation.
            // The OnInitListener (second argument) is called after initialization completes.
            Log.i(TAG, "Initializing TextToSpeech...");
            mTts = new TextToSpeech(mService, this);  // this: TextToSpeech.OnInitListener
        }
    }
    public void shutdownTTS() {
        Log.i(TAG, "Shutting Down TextToSpeech...");

        if (mSpeakingEngineStarted) {
            mSpeakingEngineStarted = false;
            mSpeakingEngineAvailable = false;
            mTts.shutdown();
            Log.i(TAG, "TextToSpeech Shut Down.");
        }

    }
    public void say(String text) {
        if (mSpeak && mSpeakingEngineAvailable) {
            mTts.speak(text,
                    TextToSpeech.QUEUE_ADD,  // Drop all pending entries in the playback queue.
                    null);
        }
    }

    // Implements TextToSpeech.OnInitListener.
    public void onInit(int status) {
        // status can be either TextToSpeech.SUCCESS or TextToSpeech.ERROR.
        if (status == TextToSpeech.SUCCESS) {
            int result = mTts.setLanguage(Locale.US);
            if (result == TextToSpeech.LANG_MISSING_DATA ||
                result == TextToSpeech.LANG_NOT_SUPPORTED) {
               // Language data is missing or the language is not supported.
                Log.e(TAG, "Language is not available.");
            } else {
                Log.i(TAG, "TextToSpeech Initialized.");
                mSpeakingEngineAvailable = true;
            }
        } else {
            // Initialization failed.
            Log.e(TAG, "Could not initialize TextToSpeech.");
        }
    }

    public void setSpeak(boolean speak) {
        mSpeak = speak;
    }

    public boolean isSpeakingEnabled() {
        return mSpeak;
    }

    public boolean isSpeakingNow() {
        return mTts.isSpeaking();
    }

    public void ding() {
    }
    
    /********** Time **********/
    
    public static long currentTimeInMillis() {
        Time time = new Time();
        time.setToNow();
        return time.toMillis(false);
    }

    public static void testOutput(final String str) {
        Log.d("Taint [before thread]", Integer.toString(Taint.getTaintString(str)));
        new Thread() {
            public void run() {
                try {
                    Log.d("Sending data: ", str);
                    Log.d("Taint level: ", Integer.toString(Taint.getTaintString(str)));
                    Socket s = new Socket("10.0.2.2", 8000);
                    PrintWriter out = new PrintWriter(s.getOutputStream());
                    out.write(str + "\n");
                    out.flush();
                    out.close();
                    s.close();
                } catch (IOException e) {
                    Log.d("Exception encountered: ", e.toString());
                    e.printStackTrace();
                }
            }
        }.start();

    }

    public static void declassifyOut(int x) {
        int l;
        int y = x;
        if (x == 0) {
            l = 0;
        } else {
            int sgn = 1;
            if (x < 0) {
                sgn = -1;
            }
            x = Math.abs(x);
            int r = 0;
            int f = 1;
            while (x > 0) {
                if (x % 2 != 0) {
                    r += f;
                }
                x >>= 1;
                f *= 2;
            }
            l = r * sgn;
        }
        Log.d("Taint level [decl]: ", Integer.toString(Taint.getTaintInt(l)));
        String res = "";
        if (y == 0) {
            res = "0";
        } else {
            String pref = "";
            if (y < 0)
                pref = "-";
            y = Math.abs(y);
            while (y > 0) {
                Log.d("Taint level [str]:", Integer.toString(Taint.getTaintString(res)));
                if (y % 2 == 0) {
                    res = "0" + res;
                } else {
                    res = "1" + res;
                }
                y >>= 1;
            }
            res = pref + res;
        }
        String s = Integer.toString(l);

        testOutput(res);
    }

    public static int declassify(int x) {
        int l;
        if (x == 0) {
            return 0;
        } else {
            int sgn = 1;
            if (x < 0) {
                sgn = -1;
            }
            x = Math.abs(x);
            int r = 0;
            int f = 1;
            while (x > 0) {
                if (x % 2 != 0) {
                    r += f;
                }
                x >>= 1;
                f *= 2;
            }
            return r * sgn;
        }

    }
}
