package complicit

import scala.annotation.tailrec
import scala.collection.JavaConverters._
import scala.collection.mutable.{Buffer, Queue, Map}
import scala.collection.Set
import scala.collection.mutable.HashSet
import scala.util.Random;
import util.control.Breaks._
// import scalaz.{Value => ZValue, _}, Scalaz.{callableMonad => _, _}

import soot.{Unit => SootUnit, util => _, _}
import soot.toolkits.graph._
import soot.jimple._
import soot.util.cfgcmd.CFGToDotGraph
import soot.util.dot._

import Lattice.Level.{Value => _, _}

class TaintTransformer extends BodyTransformer {

  var cfg: UnitGraph = null
  var body: Body = null
  var lg: soot.javaToJimple.LocalGenerator = null
  var oldMeth: SootMethod = null
  var locPC: Local = null
  var firstNonIds: Set[SootUnit] = null
  val gt = GlobalTransform.v
  val taintType = IntType.v

  var wrapperClasses: Set[SootClass] = null
  var tailStart: Map[SootUnit, SootUnit] = Map()

  // this is a bit unclean:
  var latestUnit: SootUnit = null
  var insertDirectly: Boolean = true
  var emittedCode: Seq[SootUnit] = Seq()

  protected override def internalTransform(body: Body, phase: String,
    options: java.util.Map[String, String]): Unit = {
    // Needs to be initialized here, since soot is not initialized
    // sufficiently before.
    if (wrapperClasses == null) {
      wrapperClasses = Set("java.lang.Float", "java.lang.Integer",
        "java.lang.Double", "java.lang.Boolean") map (Scene.v.loadClassAndSupport(_))
    }
    this.body = body
    val clasName = body.getMethod.getDeclaringClass.toString
    if (Main.options.ignoreClasses.contains(clasName)) {
      Util.indented { Util.debug(s"Ignoring class $clasName") }
      return
    }
    if (!GlobalTransform.v.isApplicationMethod(body.getMethod)) {
      Util.indented {
        // Util.debug(s"Skipping non-application method ${body.getMethod}\n")
      }
      return
    }
    if (isOutputMethod(body.getMethod) != None ||
      isInputMethod(body.getMethod) != None || gt.isHelperMethod(body.getMethod)) {
      return
    }
    if (clasName.contains("TaintDroidSinkUtil"))
      return
    Util.debug(s"Analyzing method ${body.getMethod} " ++
      s"[${body.getMethod.getDeclaringClass()}]")
    lg = new soot.javaToJimple.LocalGenerator(body)
    Util.overview.logMethod(body.getMethod, "original", body)
    oldMeth = body.getMethod
    Util.indented {
      handleBody(body)
      Util.overview.logMethod(oldMeth, "transformed", body)
      // Util.debug(s"Validating body for ${body.getMethod}")
      // val errList: java.util.List[soot.validation.ValidationException] =
      //   new java.util.Vector()
      // body.validate(errList)
      // if (!errList.isEmpty) {
      //   errList.asScala foreach (e => Util.warn(s"Validation error ${e}"))
      //   Util.die(s"Validation failed for body of ${body.getMethod}")
      // }
      Util.debug(s"Finished processing body for ${body.getMethod}\n")
    }
    Util.setIndent(0)
  }

  def isEntrypoint(meth: SootMethod): Boolean = {
    // quick-and-dirty heuristic for detecting android event
    // handlers
    val eventHandlers = Seq("onCreateOptionsMenu",
      "onOptionsItemSelected", "onCreate")
    meth.isEntryMethod || eventHandlers.exists (meth.toString.contains(_))
  }

  def recomputeCFG(): Unit = {
    cfg = new ExceptionalUnitGraph(body)
  }

  def handleBody(body : Body): Unit = {

    val units: Queue[SootUnit] = Queue()
    var seen: Set[SootUnit] = Set()
    // cfg = new BriefUnitGraph(body)
    cfg = new ExceptionalUnitGraph(body)
    sanitizeCFG()
    locPC = lg.generateLocal(GlobalTransform.v.pcField.getType)
    firstNonIds = findFirstNonIds(cfg.getHeads.asScala)
    firstNonIds foreach { nonId =>
      val locPCCopy = Jimple.v.newAssignStmt(locPC,
        Jimple.v.newStaticFieldRef(GlobalTransform.v.pcField.makeRef))
      body.getUnits.insertBefore(locPCCopy, nonId)
    }
    // TODO: sync with global PC before/after method calls.
    Util.debug("Iterating over units:")
    units.enqueue(cfg.getHeads.asScala: _*)
    val allUnits: Set[SootUnit] = body.getUnits.asScala.toSet
    cfg.getTails.asScala foreach { t => tailStart += ((t, t)) }
    while (!units.isEmpty) {
      breakable {
        val curUnit = units.dequeue()
        latestUnit = curUnit
        if (seen contains curUnit) {
          Util.debug(s"Already seen: $curUnit")
          break
        }
        Util.debug(s"Analyzing unit: $curUnit [${curUnit.getClass}]")
        seen += curUnit
        Util.incIndent()
        handleUnit(curUnit)
        Util.decIndent()
        cfg.getSuccsOf(curUnit).asScala foreach { u =>
          if ((! (seen contains u)) && (allUnits contains u))
            units.enqueue(u)
        }
      }
    }
    // sync with global PC at each exit point:
    cfg.getTails.asScala foreach { exitUnit =>
      val copyToGlobalPC = Jimple.v.newAssignStmt(
        Jimple.v.newStaticFieldRef(GlobalTransform.v.pcField.makeRef),
        locPC)
      body.getUnits.insertBefore(copyToGlobalPC, exitUnit)
    }
  }

  def emit(stms: Seq[SootUnit]): Unit = {
    if (insertDirectly) {
      body.getUnits.insertAfter(stms.asJava, latestUnit)
      latestUnit = stms.last
    } else {
      emittedCode = emittedCode ++ stms
    }
  }

  def emit(stm: SootUnit): Unit = emit(Seq(stm))

  def catchEmittedCode(act: => Unit): Seq[SootUnit] = {
    insertDirectly = false
    emittedCode = Seq()
    act
    insertDirectly = true
    emittedCode
  }

  def findFirstNonIds(starts: Seq[SootUnit]): Set[SootUnit] = {
    (starts flatMap findFirstNonId).toSet
  }

  def findFirstNonId(u: SootUnit): Set[SootUnit] = {
    u match {
      case id: IdentityStmt =>
        val next = cfg.getSuccsOf(u).asScala
        if (next.length == 0) {
          Set(u)
        } else {
          findFirstNonIds(next)
        }
      case _ => Set(u)
    }
  }

  def handleUnit(unit: SootUnit) : Unit = {
    val clasPattern = """class soot\.jimple\.internal\.(.*)$""".r
    val clas = clasPattern replaceAllIn
      (unit.getClass.toString, _ match { case clasPattern(x) => x })
    val lg = new soot.javaToJimple.LocalGenerator(body)
    unit match {
      case a: AssignStmt => handleAssign(a)
      case br: IfStmt => handleIf(br)
      case inv: InvokeStmt =>
        val meth = inv.getInvokeExpr.getMethod
        isOutputMethod(inv.getInvokeExpr.getMethod) match {
          case Some(l) => emit(Jimple.v.newInvokeStmt(
            Jimple.v.newStaticInvokeExpr(gt.outMethod.makeRef, locPC)))
          case _ =>
        }
        if (gt.isDeclassifyMethod(meth)) {
          // set global pc to  0
          val lowerPCStmt = Jimple.v.newAssignStmt(
            Jimple.v.newStaticFieldRef(gt.pcField.makeRef),
            IntConstant.v(0))
          body.getUnits.insertBefore(lowerPCStmt, inv)
          inv.redirectJumpsToThisTo(lowerPCStmt)
          // restore after exiting method
          body.getUnits.insertAfter(
            Jimple.v.newAssignStmt(
              Jimple.v.newStaticFieldRef(gt.pcField.makeRef),
              locPC), inv)
        } else {
          val globPCStmt = Jimple.v.newAssignStmt(Jimple.v.newStaticFieldRef(gt.pcField.makeRef),
            locPC)
          body.getUnits.insertBefore(globPCStmt, inv)
          inv.redirectJumpsToThisTo(globPCStmt)
        }
      case lswitch: LookupSwitchStmt => handleLookupSwitch(lswitch)
      case _ => Util.warn(s"Unhandled statement type: $unit (${unit.getClass}")
    }
  }

  def withSavedPC(i: Stmt, act: => Unit): Unit = {
    val (bakPCStmt, mkRestoreStms) = savePCStmts()
    cfg = new ExceptionalUnitGraph(body)
    val postDoms = new MHGPostDominatorsFinder(cfg)
    val dom = postDoms.getImmediateDominator(i)
    if (dom == null) {
      Util.warn("No post dominator found")
      // this means that we exit the method before being able to restore the pc
      // put a restore statement before each method exit:
      cfg.getTails.asScala foreach { t =>
        val ts = tailStart.getOrElse(t, Util.die(s"No tail start for $t"))
        val restoreStms = mkRestoreStms()
        body.getUnits.insertBefore(restoreStms.asJava, ts)
        t.redirectJumpsToThisTo(restoreStms.head)
        tailStart += ((t, restoreStms.head))
      }
    } else {
      val restoreStms = mkRestoreStms()
      body.getUnits.insertBefore(restoreStms.asJava, dom)
      dom.redirectJumpsToThisTo(restoreStms.head)
    }
    body.getUnits.insertBefore(bakPCStmt, i)
    act
    i.redirectJumpsToThisTo(bakPCStmt)
  }

  def savePCAndInsert(i: Stmt, toInsert: Seq[SootUnit]) =
    withSavedPC(i, body.getUnits.insertBefore(toInsert.asJava, i))

  def handleLookupSwitch(lswitch: LookupSwitchStmt): Unit =
    savePCAndInsert(lswitch, catchEmittedCode { addTaint(locPC, lswitch.getKey) })

  def handleIf(i: IfStmt): Unit = savePCAndInsert(i, addCondToPC(i.getCondition))

  def addCondToPC(cond: Value): Seq[SootUnit] = {
    cond match {
      case _ : Constant => List()
      case binE : BinopExpr => {
        addCondToPC(binE.getOp1) ++
          addCondToPC(binE.getOp2)
      }
      case _ =>
        catchEmittedCode(addTaint(locPC, cond))
    }
  }

  /** Generate statements to save and restore the PC around
    an if statement.
  *
  *  Ugly workaround: Since Java programs can result
  *  in nested blocks where the one branch and the outer
  *  if share an immediate post-dominator, we only restore
  *  the pc if we come "from the right block". This is
  *  done by initializing the fresh variables to random
  *  integers and only restoring if the value of the variable
  *  no longer matches this integer at the dominator. This
  *  is not a security concern since, in the worst case,
  *  the attacker forces the PC to be restored to the value
  *  of a nested block, the level of which is lower-bounded by
  *  the level of surrounding if anyway. However, this
  *  has a very low chance of introducing random false positives.
  *
  * @return A statement to prepend to the if and a list of statements to be
  * 	inserted before the dominator.
  */
  def savePCStmts(): (AssignStmt, (Unit => List[SootUnit])) = {
    val fv = lg.generateLocal(locPC.getType)
    val saveStmt = Jimple.v.newAssignStmt(fv, locPC)
    val initVal = IntConstant.v(new Random().nextInt())
    firstNonIds foreach
      (body.getUnits.insertBefore(Jimple.v.newAssignStmt(fv, initVal), _))
    /* Inserts roughly this:
     T method(..) {
       fv1 := N // N is a random constant
       ..
       fv1 := pc
       pc := pc + e
       if e goto l
       ...
       l:
       ...
       // control-flow join point:
       if fv1 == N goto end
       pc := fv1
       end:
       nop // this could be optimized
       ...
     }
     */
    // TODO: avoid superfluous nop
    val stringClass = Scene.v.getClasses.asScala find (_.toString == "java.lang.String")
    stringClass match {
      case Some(sc) => {
        (saveStmt, (_ => {
          val nopBox = Jimple.v.newStmtBox(Jimple.v.newNopStmt())
          val restore1 = Jimple.v.newIfStmt(Jimple.v.newEqExpr(fv,
            initVal), nopBox)
          val restore2 = Jimple.v.newAssignStmt(locPC, fv)
          List(restore1, restore2, nopBox.getUnit)
        }))
      }
      case None => Util.die("String class not found")
    }
  }

  def ensureLocal(x: Value): Local = x match {
    case l: Local => l
    case _ =>
      val tmpLoc = lg.generateLocal(x.getType)
      emit(Jimple.v.newAssignStmt(tmpLoc, x))
      tmpLoc
  }

  def sanitizeCFG(): Unit = {
    // Perhaps we return directly from at least one branch.
    // In that case, we must have multiple exit points in the CFG.
    val retType = body.getMethod.getReturnType
    val isVoid = (retType == VoidType.v)
    if (cfg.getTails.size > 1 &&
        (cfg.getTails.asScala forall
          (u => u.isInstanceOf[ReturnStmt] || u.isInstanceOf[ReturnVoidStmt]))) {
      Util.debug("Found CFG with multiple exits, fixing..")
      // Insert a new local to hold the return value (unless void)
      var retLocal: Local = null      // Recompute cfg and dominators
      // cfg = new BriefUnitGraph(body)
      // domFinder = new MHGPostDominatorsFinder(cfg)

      if (!isVoid) {
        retLocal = lg.generateLocal(retType)
      }
      // Insert new block from which to return
      val newRet: Stmt = if (isVoid) Jimple.v.newReturnVoidStmt()
        else Jimple.v.newReturnStmt(retLocal)
      body.getUnits.addLast(newRet)
      for (u <- cfg.getTails.asScala) {
        u match {
          case retU: ReturnStmt => {
            val jmpToEnd = Jimple.v.newGotoStmt(newRet)
            if (!isVoid) {
              val saveRet = Jimple.v.newAssignStmt(retLocal, retU.getOp)
              u.redirectJumpsToThisTo(saveRet)
              body.getUnits.insertBefore(saveRet, retU)
              body.getUnits.remove(retU)
              body.getUnits.insertAfter(jmpToEnd, saveRet)
            } else {
              u.redirectJumpsToThisTo(jmpToEnd)
              body.getUnits.insertBefore(jmpToEnd, retU)
              body.getUnits.remove(retU)
            }
          }
          case retU: ReturnVoidStmt =>
            val jmpToEnd = Jimple.v.newGotoStmt(newRet)
            u.redirectJumpsToThisTo(jmpToEnd)
            body.getUnits.insertBefore(jmpToEnd, retU)
            body.getUnits.remove(retU)
          case _ => Util.die("Non-return in tail of CFG")
        }
      }
      recomputeCFG()
    }
  }

  def upgradeTaint(x: Value, taint: Local): Unit = {
    gt.upgradeMethods.get(x.getType) match {
      case Some(upgMeth) =>
        val xLoc = ensureLocal(x)
        if (upgMeth.getReturnType.equals(x.getType)) {
          val invokeStmt = Jimple.v.newAssignStmt(xLoc,
            Jimple.v.newStaticInvokeExpr(upgMeth.makeRef, xLoc, taint))
          emit(invokeStmt)
          if (!xLoc.equals(x)) {
            emit(Jimple.v.newAssignStmt(x, xLoc))
          }
        }
        val invokeStmt: SootUnit = Jimple.v.newInvokeStmt(
            Jimple.v.newStaticInvokeExpr(upgMeth.makeRef,
              xLoc, taint))
        emit(invokeStmt)
      case _ =>
        Util.warn(s"Unhandled non-taint-carrying type ${x.getType}")
    }
  }

  def getTaint(x: Value): Local = {
    val taintLoc = lg.generateLocal(taintType)
    gt.getTaintMethods.get(x.getType) match {
      case Some(getMeth) =>
        val xLoc = ensureLocal(x)
        val assignResult = Jimple.v.newAssignStmt(taintLoc,
          Jimple.v.newStaticInvokeExpr(getMeth.makeRef, xLoc))
        emit(assignResult)
      case None =>
        Util.warn(s"Unhandled non-taint-carrying type ${x.getType}")
        val assign0: SootUnit = Jimple.v.newAssignStmt(taintLoc, IntConstant.v(0))
        emit(assign0)
    }
    taintLoc
  }

  // Add the taint of value t to that of x
  def addTaint(x: Value, t: Value): Unit = upgradeTaint(x, getTaint(t))

  def handleAssign(a: AssignStmt): Unit = {
    // add a workaround for pedometer example, to mark
    // sensor events as a high source:
    // TODO: avoid recomputing this
    val sensorClass = Scene.v.loadClassAndSupport("android.hardware.Sensor")
    val getTypeMeth = sensorClass.getMethod("getType", Seq().asJava, IntType.v)
    a.getRightOp match {
      case inv: InvokeExpr
          if inv.getMethod.equals(getTypeMeth) =>
        // raise taint unconditionally of PC:
        val constLoc = lg.generateLocal(IntType.v)
        emit(Jimple.v.newAssignStmt(constLoc, IntConstant.v(4242)))
        upgradeTaint(a.getLeftOp, constLoc)
      case _ =>  addTaint(a.getLeftOp(), locPC)
    }
  }

  def isInputMethod(m: SootMethod): Option[Level] = {
    val methName = m.toString
    val inputPatterns = Seq("getSecret", "getDeviceId",
      "getLatitude", "getLongitude", "getSimSerialNumber")
    if (inputPatterns.exists (methName.contains(_))) {
      Some(H)
    } else {
      None
    }
  }

  def isOutputMethod(m: SootMethod): Option[Level] = {
    val methName = m.toString
    val outputPatterns = Seq("println", "sendTextMessage", "testOutput")

    if (outputPatterns.exists (methName.contains(_))) {
      Some(L)
    } else {
      None
    }
  }

  def isInput(expr: InvokeExpr): Option[Level] =
    isInputMethod(expr.getMethod)

  def isOutput(expr: InvokeExpr): Option[Level] =
    isOutputMethod(expr.getMethod)

  def isApplicationMethod(m: SootMethod): Boolean =
    Scene.v.getApplicationClasses.asScala.toList.contains(m.getClass())

}

object TaintTransformer {
  private val instance: TaintTransformer = new TaintTransformer()
  def v: TaintTransformer = return instance

}
